@@ -44,9 +44,13 @@
 #ifdef HAVE_LINUX_SECUREBITS_H
 #include <linux/securebits.h>
 #endif
+#ifdef HAVE_LINUX_MAGIC_H
+#include <sys/vfs.h>
+#include <linux/magic.h>
+#endif
 
 # define hidden __attribute__ ((visibility ("hidden")))
-int last_cap hidden = -1;
+unsigned int last_cap hidden = 0;
 /*
  * Some milestones of when things became available:
  * 2.6.24 kernel	XATTR_NAME_CAPS
@@ -69,7 +73,7 @@ static void update_ambient_set(capng_act_t action, unsigned int capability,
 // Local defines
 #define MASK(x) (1U << (x))
 #ifdef PR_CAPBSET_DROP
-#define UPPER_MASK ~(unsigned)((~0U)<<(last_cap-31))
+#define UPPER_MASK ~((~0U)<<(last_cap-31))
 #else
 // For v1 systems UPPER_MASK will never be used
 #define UPPER_MASK (unsigned)(~0U)
@@ -77,7 +81,7 @@ static void update_ambient_set(capng_act_t action, unsigned int capability,
 
 // Re-define cap_valid so its uniform between V1 and V3
 #undef cap_valid
-#define cap_valid(x) ((x) <= (unsigned int)last_cap)
+#define cap_valid(x) ((x) <= last_cap)
 
 // If we don't have the xattr library, then we can't
 // compile-in file system capabilities
@@ -185,12 +189,63 @@ static void deinit(void)
 	m.state = CAPNG_NEW;
 }
 
+static inline int test_cap(unsigned int cap)
+{
+	// prctl returns 0 or 1 for valid caps, -1 otherwise
+	return prctl(PR_CAPBSET_READ, cap) >= 0;
+}
+
+// The maximum cap value is determined by VFS_CAP_U32
+#define MAX_CAP_VALUE (VFS_CAP_U32 * sizeof(__le32) * 8)
+
 static void init_lib(void) __attribute__ ((constructor));
 static void init_lib(void)
 {
 #ifdef HAVE_PTHREAD_H
 	pthread_atfork(NULL, NULL, deinit);
 #endif
+	// Detect last cap
+	if (last_cap == 0) {
+		int fd;
+
+		// Try to read last cap from procfs
+		fd = open("/proc/sys/kernel/cap_last_cap", O_RDONLY);
+		if (fd >= 0) {
+#ifdef HAVE_LINUX_MAGIC_H
+			struct statfs st;
+			// Bail out if procfs is invalid or fstatfs fails
+			if (fstatfs(fd, &st) || st.f_type != PROC_SUPER_MAGIC)
+				goto fail;
+#endif
+			char buf[8];
+			int num = read(fd, buf, sizeof(buf) - 1);
+			if (num > 0) {
+				buf[num] = 0;
+				errno = 0;
+				unsigned int val = strtoul(buf, NULL, 10);
+				if (errno == 0)
+					last_cap = val;
+			}
+fail:
+			close(fd);
+		}
+		// Run a binary search over capabilities
+		if (last_cap == 0) {
+			// starting with last_cap=MAX_CAP_VALUE means we always know
+			// that cap1 is invalid after the first iteration
+			last_cap = MAX_CAP_VALUE;
+			unsigned int cap0 = 0, cap1 = MAX_CAP_VALUE;
+
+			while (cap0 < last_cap) {
+				if (test_cap(last_cap))
+					cap0 = last_cap;
+				else
+					cap1 = last_cap;
+
+				last_cap = (cap0 + cap1) / 2U;
+			}
+		}
+	}
 }
 
 static void init(void)
@@ -222,31 +277,6 @@ static void init(void)
 #else
 	m.hdr.pid = (unsigned)getpid();
 #endif
-	// Detect last cap
-	if (last_cap == -1) {
-		int fd;
-
-		fd = open("/proc/sys/kernel/cap_last_cap", O_RDONLY);
-		if (fd == -1) {
-			if (errno != ENOENT) {
-				m.state = CAPNG_ERROR;
-				return;
-			}
-		} else {
-			char buf[8];
-			int num = read(fd, buf, sizeof(buf) - 1);
-			if (num > 0) {
-				buf[num] = 0;
-				errno = 0;
-				int val = strtoul(buf, NULL, 10);
-				if (errno == 0)
-					last_cap = val;
-			}
-			close(fd);
-		}
-		if (last_cap == -1)
-			last_cap = CAP_LAST_CAP;
-	}
 	m.rootid = CAPNG_UNSET_ROOTID;
 	m.state = CAPNG_ALLOCATED;
 }
@@ -372,8 +402,8 @@ static int get_bounding_set(void)
 			fclose(f);
 			return 0;
 		}
+		// Didn't find bounding set, fall through and try prctl way
 		fclose(f);
-		return -1;
 	}
 	// Might be in a container with no procfs - do it the hard way
 	memset(m.bounds, 0, sizeof(m.bounds));
@@ -418,7 +448,7 @@ static int get_ambient_set(void)
 			return 0;
 		}
 		fclose(f);
-		return -1;
+		// Didn't find ambient set, fall through and try prctl way
 	}
 	// Might be in a container with no procfs - do it the hard way
 	memset(m.ambient, 0, sizeof(m.ambient));
@@ -685,7 +715,7 @@ int capng_updatev(capng_act_t action, capng_type_t type,
 
 int capng_apply(capng_select_t set)
 {
-	int rc = -1;
+	int rc = 0;
 
 	// Before updating, we expect that the data is initialized to something
 	if (m.state < CAPNG_INIT)
@@ -699,48 +729,80 @@ int capng_apply(capng_select_t set)
 		if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP)) {
 			unsigned int i;
 			memcpy(&m, &state, sizeof(m)); /* restore state */
-			rc = 0;
-			for (i=0; i <= last_cap && rc == 0; i++)
+			for (i=0; i <= last_cap; i++) {
 				if (capng_have_capability(CAPNG_BOUNDING_SET,
-								 i) == 0)
-					rc = prctl(PR_CAPBSET_DROP, i, 0, 0, 0);
-			if (rc == 0)
-				m.state = CAPNG_APPLIED;
-		} else
+								 i) == 0) {
+				    if (prctl(PR_CAPBSET_DROP, i, 0, 0, 0) <0) {
+					rc = -2;
+					goto try_caps;
+				    }
+				}
+			}
+			m.state = CAPNG_APPLIED;
+			if (get_bounding_set() < 0) {
+				rc = -3;
+				goto try_caps;
+			}
+		} else {
 			memcpy(&m, &state, sizeof(m)); /* restore state */
-#else
-		rc = 0;
+			rc = -4;
+			goto try_caps;
+		}
 #endif
 	}
+
+	// Try caps is here so that if someone had SELECT_BOTH and we blew up
+	// doing the bounding set, we at least try to set any capabilities
+	// before returning in case the caller also doesn't bother checking
+	// the return code.
+try_caps:
+	if (set & CAPNG_SELECT_CAPS) {
+		if (capset((cap_user_header_t)&m.hdr,
+				(cap_user_data_t)&m.data) == 0)
+			m.state = CAPNG_APPLIED;
+		else
+			rc = -5;
+	}
+
+	// Most programs do not and should not mess with ambient capabilities.
+	// Instead of returning here if rc is set, we'll let it try to
+	// do something with ambient capabilities in hopes that it's lowering
+	// capabilities. Again, this is for people that don't check their
+	// return codes.
+	//
+	// Do ambient last so that inheritable and permitted are set by the
+	// time we get here.
 	if (set & CAPNG_SELECT_AMBIENT) {
 #ifdef PR_CAP_AMBIENT
-		unsigned int i;
 		if (capng_have_capabilities(CAPNG_SELECT_AMBIENT) ==
 								CAPNG_NONE) {
-			rc = prctl(PR_CAP_AMBIENT,
-					   PR_CAP_AMBIENT_CLEAR_ALL, 0, 0, 0);
+			if (prctl(PR_CAP_AMBIENT,
+				   PR_CAP_AMBIENT_CLEAR_ALL, 0, 0, 0) < 0) {
+				rc = -6;
+				goto out;
+			}
 		} else {
+			unsigned int i;
+
 			// Clear them all
-			rc = prctl(PR_CAP_AMBIENT,
-				   PR_CAP_AMBIENT_CLEAR_ALL, 0, 0, 0);
-			for (i=0; i <= last_cap && rc == 0; i++) {
+			if (prctl(PR_CAP_AMBIENT,
+				   PR_CAP_AMBIENT_CLEAR_ALL, 0, 0, 0) < 0) {
+				rc = -7;
+				goto out;
+			}
+			for (i=0; i <= last_cap; i++) {
 				if (capng_have_capability(CAPNG_AMBIENT, i))
-					rc = prctl(PR_CAP_AMBIENT,
-						PR_CAP_AMBIENT_RAISE, i, 0, 0);
+					if (prctl(PR_CAP_AMBIENT,
+					    PR_CAP_AMBIENT_RAISE, i, 0, 0) < 0){
+						rc = -8;
+						goto out;
+					}
 			}
 		}
-		if (rc == 0)
-			m.state = CAPNG_APPLIED;
-#else
-		rc = 0;
+		m.state = CAPNG_APPLIED;
 #endif
 	}
-	if (set & CAPNG_SELECT_CAPS) {
-		rc = capset((cap_user_header_t)&m.hdr,
-				(cap_user_data_t)&m.data);
-		if (rc == 0)
-			m.state = CAPNG_APPLIED;
-	}
+out:
 	return rc;
 }
 
@@ -1119,7 +1181,7 @@ int capng_have_capability(capng_type_t which, unsigned int capability)
 
 	// If we still don't have anything, error out
 	if (m.state < CAPNG_INIT)
-		return CAPNG_FAIL;
+		return 0;
 	if (m.cap_ver == 1 && capability > 31)
 		return 0;
 	if (!cap_valid(capability))
@@ -1258,7 +1320,7 @@ char *capng_print_caps_numeric(capng_print_t where, capng_select_t set)
 
 char *capng_print_caps_text(capng_print_t where, capng_type_t which)
 {
-	int i, once = 0, cnt = 0;
+	unsigned int i, once = 0, cnt = 0;
 	char *ptr = NULL;
 
 	if (m.state < CAPNG_INIT)
